import numpy as np
from mpl_toolkits.mplot3d import Axes3D
import math
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import gym
from gym import spaces
import matplotlib.patches as patches
import time
import cv2

class Env3DStatic(gym.Env):
    def __init__(self, args):

        self.step_size = 1

        self.plan_width = 20  # X Axis
        self.plan_height = 20  # Y Axis
        self.plan_length = 10  # Z Axis
        self.z = 6

        self.plan_choose = (args.plan_choose if args.plan_choose is not None else 0)

        self.environment_memory = None
        self.count_brick = None
        self.brick_memory = None
        self.HALF_WINDOW_SIZE = 3
        self.environment_width = self.plan_width + 2 * self.HALF_WINDOW_SIZE
        self.environment_height = self.plan_height + 2 * self.HALF_WINDOW_SIZE
        self.environment_length = self.plan_length + 2 * self.HALF_WINDOW_SIZE

        self.start = None
        self.position_memory = None
        self.observation = None
        self.count_step = 0
        self.total_step = 1300
        self.plan, self.total_brick = self.create_plan()
        self.input_plan = None  # outer_use

        self.one_hot = None
        self.check = []
        self.action_dim = 8
        # up, down, left, right,
        # up_brick, down_brick, left_brick, right_brick
        self.state_dim = (2 * self.HALF_WINDOW_SIZE + 1) ** 2 + 2
        self.blank_size = 2  # use 0, 2, 4
        self.start = None
        self.window = args.history_length
        self.uniform_step = args.uniform_step

        self.features = 51 - 1

    def set_plan_choose(self, plan_choose):
        self.plan_choose = plan_choose

    def get_features(self):
        return self.features

    def action_space(self):
        return self.action_dim

    def create_plan(self):
        out_radius = 7
        if self.plan_choose == 0:
            in_radius = 0
        elif self.plan_choose == 1:
            out_radius = 8
            in_radius = 7
        else:
            raise ValueError('0: Dense circle, 1: Sparse circle')

        plan = np.zeros((self.environment_height, self.environment_width))
        a = np.array([12.5, 12.5])
        circle = patches.CirclePolygon(a, out_radius)
        circle2 = patches.CirclePolygon(a, in_radius)
        for i in range(len(plan)):
            for j in range(len(plan[0])):
                a = np.array([i, j])
                if circle.contains_point(a) and not circle2.contains_point(a):
                    plan[i][j] = 1

            total_area = sum(sum(plan))
        plan = plan * self.z
        total_area = total_area * self.z
        return plan, total_area

    def reset(self):
        self.plan, self.total_brick = self.create_plan()
        self.input_plan = self.plan[self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_height,
                          self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_width]
        self.environment_memory = np.zeros((self.environment_height, self.environment_width))
        self.environment_memory[:, :self.HALF_WINDOW_SIZE] = -1
        self.environment_memory[:, -self.HALF_WINDOW_SIZE:] = -1
        self.environment_memory[:self.HALF_WINDOW_SIZE, :] = -1
        self.environment_memory[-self.HALF_WINDOW_SIZE:, :] = -1
        self.check = []
        self.count_brick = 0
        self.step_size = 1
        self.position_memory = []
        self.observation = None
        self.count_step = 0
        initial_position = [self.HALF_WINDOW_SIZE, self.HALF_WINDOW_SIZE]

        self.position_memory.append(initial_position)
        return np.hstack(
            (self.observation_(initial_position), np.array([[self.count_brick]]), np.array([[self.count_step]])))

    def check_sur(self, position):
        # 0 left, 1 right, 2 up, 3 down
        check = [0, 0, 0, 0, 0, 0, 0, 0]

        observation = [self.environment_memory[position[0], position[1] - 1],
                       self.environment_memory[position[0], position[1] + 1],
                       self.environment_memory[position[0] + 1, position[1]],
                       self.environment_memory[position[0] - 1, position[1]]]
        for i in range(len(observation)):
            if observation[i] == -1:
                check[i] = 1
                check[i + 4] = 1
            elif observation[i] > 0:
                check[i] = 1
        return check

    def move_step(self, position, action, step_size):
        move = 0
        if action == 0:
            for i in range(step_size):
                observation = self.environment_memory[position[0], position[1] - i - 1]
                if observation == 0:
                    move += 1
                else:
                    break
        elif action == 1:
            for i in range(step_size):
                observation = self.environment_memory[position[0], position[1] + i + 1]
                if observation == 0:
                    move += 1
                else:
                    break
        elif action == 2:
            for i in range(step_size):
                observation = self.environment_memory[position[0] + i + 1, position[1]]
                if observation == 0:
                    move += 1
                else:
                    break
        elif action == 3:
            for i in range(step_size):
                observation = self.environment_memory[position[0] - i - 1, position[1]]
                if observation == 0:
                    move += 1
                else:
                    break
        return move

    def observation_(self, position):
        observation = self.environment_memory[
                      position[0] - self.HALF_WINDOW_SIZE:position[0] + self.HALF_WINDOW_SIZE + 1,
                      position[1] - self.HALF_WINDOW_SIZE:position[1] + self.HALF_WINDOW_SIZE + 1]
        return observation.flatten().reshape(1, -1)

    def clip_position(self, position):
        if position[0] <= self.HALF_WINDOW_SIZE:
            position[0] = self.HALF_WINDOW_SIZE
        if position[1] <= self.HALF_WINDOW_SIZE:
            position[1] = self.HALF_WINDOW_SIZE
        if position[0] >= self.plan_width + self.HALF_WINDOW_SIZE - 1:
            position[0] = self.plan_width + self.HALF_WINDOW_SIZE - 1
        if position[1] >= self.plan_width + self.HALF_WINDOW_SIZE - 1:
            position[1] = self.plan_width + self.HALF_WINDOW_SIZE - 1
        return position

    def step(self, action):
        self.count_step += 1
        if self.uniform_step:
            self.step_size = 1
        else:
            self.step_size = np.random.randint(1, 4)
        # 0 left, 1 right, 2 up, 3 down, 4 left_brick, 5 right_brick, 6 up_brick, 7 down_brick

        check = self.check_sur(self.position_memory[-1])
        self.check = check
        # moving: 0 left, 1 right, 2 up, 3 down
        if action == 0 and check[0] == 0:
            step_size = self.move_step([self.position_memory[-1][0], self.position_memory[-1][1]], action,
                                       self.step_size)
            position = [self.position_memory[-1][0], self.position_memory[-1][1] - step_size]
            position = self.clip_position(position)
            self.position_memory.append(position)
        elif action == 1 and check[1] == 0:
            step_size = self.move_step([self.position_memory[-1][0], self.position_memory[-1][1]], action,
                                       self.step_size)
            position = [self.position_memory[-1][0], self.position_memory[-1][1] + step_size]
            position = self.clip_position(position)
            self.position_memory.append(position)
        elif action == 2 and check[2] == 0:
            step_size = self.move_step([self.position_memory[-1][0], self.position_memory[-1][1]], action,
                                       self.step_size)
            position = [self.position_memory[-1][0] + step_size, self.position_memory[-1][1]]
            position = self.clip_position(position)
            self.position_memory.append(position)
        elif action == 3 and check[3] == 0:
            step_size = self.move_step([self.position_memory[-1][0], self.position_memory[-1][1]], action,
                                       self.step_size)
            position = [self.position_memory[-1][0] - step_size, self.position_memory[-1][1]]
            position = self.clip_position(position)
            self.position_memory.append(position)

        # building brick: 4 left_brick, 5 right_brick, 6 up_brick, 7 down_brick
        elif action > 3:
            build_brick = True
            position = self.position_memory[-1]
            self.position_memory.append(position)
            if action == 4 and check[4] == 0:
                self.count_brick += 1
                brick_target = [position[0], position[1] - 1]
                self.environment_memory[brick_target[0], brick_target[1]] += 1.0
            elif action == 5 and check[5] == 0:
                self.count_brick += 1
                brick_target = [position[0], position[1] + 1]
                self.environment_memory[brick_target[0], brick_target[1]] += 1.0
            elif action == 6 and check[7] == 0:
                self.count_brick += 1
                brick_target = [position[0] - 1, position[1]]
                self.environment_memory[brick_target[0], brick_target[1]] += 1.0
            elif action == 7 and check[6] == 0:
                self.count_brick += 1
                brick_target = [position[0] + 1, position[1]]
                self.environment_memory[brick_target[0], brick_target[1]] += 1.0
            else:
                build_brick = False

            done = bool(self.count_brick >= self.total_brick) or check[0:4] == [1, 1, 1, 1]
            if done:
                observation = np.hstack(
                    (self.observation_(position), np.array([[self.count_brick]]), np.array([[self.count_step]])))
                reward = 0.0
                return observation, reward, done
            else:
                if build_brick:
                    reward = self.reward_check(brick_target)
                    observation = np.hstack(
                        (self.observation_(position), np.array([[self.count_brick]]), np.array([[self.count_step]])))
                    return observation, reward, done
        else:
            # illegal_move = useless_move!!
            position = self.position_memory[-1]

        done = bool(self.count_step >= self.total_step) or check[0:4] == [1, 1, 1, 1]
        observation = np.hstack(
            (self.observation_(position), np.array([[self.count_brick]]), np.array([[self.count_step]])))
        reward = 0.0
        return observation, reward, done

    def reward_check(self, position):
        if self.environment_memory[position[0], position[1]] > self.plan[position[0], position[1]]:
            reward = -1.0
        elif self.environment_memory[position[0], position[1]] == self.plan[position[0], position[1]]:
            reward = 10.0
        else:
            reward = 1.0
        return reward

    def plot_3d(self, plan, Env=False):
        if Env:
            z_max = max(max(row) for row in plan)
        else:
            z_max = 1
        X, Y, Z = [], [], []
        for i in range(int(z_max)):
            for j in range(len(plan)):
                for k in range(len(plan[0])):
                    if self.environment_length > plan[j][k] > 0:
                        X.append(k)
                        Y.append(j)
                        Z.append(plan[j][k])
            plan = plan - 1.0
        return X, Y, Z

    def _iou(self):

        plan = self.plan[self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_height,
               self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_width]

        env_memory = self.environment_memory[self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_height,
                     self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_width]

        env_memory_3d = np.zeros((self.plan_height, self.plan_width))
        for i in range(self.plan_height):
            for j in range(self.plan_width):
                if env_memory[i][j] > plan[i][j]:
                    env_memory_3d[i][j] = plan[i][j]
                else:
                    env_memory_3d[i][j] = env_memory[i][j]
        plan_area = self.total_brick
        env_area = self.count_brick
        cross_area = sum(map(sum, env_memory_3d))
        iou = cross_area / (plan_area + env_area - cross_area)
        return iou

        pass

class Env3DDynamic(gym.Env):
    def __init__(self, args):

        self.step_size = 1

        self.plan_width = 20  # X Axis
        self.plan_height = 20  # Y Axis
        self.plan_length = 10  # Z Axis
        self.z = 6

        self.plan_choose = (args.plan_choose if args.plan_choose is not None else 0)

        self.environment_memory = None
        self.count_brick = None
        self.brick_memory = None
        self.HALF_WINDOW_SIZE = 3
        self.environment_width = self.plan_width + 2 * self.HALF_WINDOW_SIZE
        self.environment_height = self.plan_height + 2 * self.HALF_WINDOW_SIZE
        self.environment_length = self.plan_length + 2 * self.HALF_WINDOW_SIZE

        self.start = None
        self.position_memory = None
        self.observation = None
        self.count_step = 0
        self.total_step = 1000
        self.plan, self.total_brick = self.create_plan()
        self.input_plan = None

        self.one_hot = None
        self.check = []
        # up, down, left, right,
        # up_brick, down_brick, left_brick, right_brick
        self.action_dim = 8

        self.state_dim = (2 * self.HALF_WINDOW_SIZE + 1) ** 2 + 2
        self.blank_size = 2  # use 0, 2, 4
        self.start = None
        self.window = args.history_length
        self.uniform_step = args.uniform_step

        self.features = 51 - 1

    def set_plan_choose(self, plan_choose):
        self.plan_choose = plan_choose

    def get_features(self):
        return self.features

    def action_space(self):
        return self.action_dim

    def create_plan(self):
        if not (self.plan_choose == 1 or self.plan_choose == 0):
            raise ValueError(' 0: Dense triangle, 1: Sparse triangle')

        plan = np.zeros((self.environment_height, self.environment_width))
        area_min = [50, 20]
        area_max = 110
        total_area = 0
        i = 0
        while total_area <= area_min[self.plan_choose] or total_area >= area_max:
            i += 1
            x = np.random.randint(0, self.plan_width, size=3)
            y = np.random.randint(0, self.plan_height, size=3)

            img_rgb = np.ones((self.plan_height, self.plan_width, 3), np.uint8) * 255
            vertices = np.array([[x[0], y[0]], [x[1], y[1]], [x[2], y[2]]], np.int32)
            pts = vertices.reshape((-1, 1, 2))
            cv2.polylines(img_rgb, [pts], isClosed=True, color=(0, 0, 0))
            if self.plan_choose == 0:
                cv2.fillPoly(img_rgb, [pts], color=(0, 0, 0))

            plan[self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_height,
            self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_width] = 1 - img_rgb[:, :, 0] / 255
            total_area = sum(sum(plan))
        plan = plan * self.z
        total_area = total_area * self.z
        return plan, total_area

    def reset(self):
        self.plan, self.total_brick = self.create_plan()
        self.input_plan = self.plan[self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_height,
                          self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_width]
        self.environment_memory = np.zeros((self.environment_height, self.environment_width))
        self.environment_memory[:, :self.HALF_WINDOW_SIZE] = -1
        self.environment_memory[:, -self.HALF_WINDOW_SIZE:] = -1
        self.environment_memory[:self.HALF_WINDOW_SIZE, :] = -1
        self.environment_memory[-self.HALF_WINDOW_SIZE:, :] = -1
        self.check = []
        self.count_brick = 0
        self.step_size = 1
        self.position_memory = []
        self.observation = None
        self.count_step = 0
        initial_position = [self.HALF_WINDOW_SIZE, self.HALF_WINDOW_SIZE]

        self.position_memory.append(initial_position)
        return np.hstack(
            (self.observation_(initial_position), np.array([[self.count_brick]]), np.array([[self.count_step]])))

    def check_sur(self, position):
        # 0 left, 1 right, 2 up, 3 down
        check = [0, 0, 0, 0, 0, 0, 0, 0]

        observation = [self.environment_memory[position[0], position[1] - 1],
                       self.environment_memory[position[0], position[1] + 1],
                       self.environment_memory[position[0] + 1, position[1]],
                       self.environment_memory[position[0] - 1, position[1]]]
        for i in range(len(observation)):
            if observation[i] == -1:
                check[i] = 1
                check[i + 4] = 1
            elif observation[i] > 0:
                check[i] = 1
        return check

    def move_step(self, position, action, step_size):
        move = 0
        if action == 0:
            for i in range(step_size):
                observation = self.environment_memory[position[0], position[1] - i - 1]
                if observation == 0:
                    move += 1
                else:
                    break
        elif action == 1:
            for i in range(step_size):
                observation = self.environment_memory[position[0], position[1] + i + 1]
                if observation == 0:
                    move += 1
                else:
                    break
        elif action == 2:
            for i in range(step_size):
                observation = self.environment_memory[position[0] + i + 1, position[1]]
                if observation == 0:
                    move += 1
                else:
                    break
        elif action == 3:
            for i in range(step_size):
                observation = self.environment_memory[position[0] - i - 1, position[1]]
                if observation == 0:
                    move += 1
                else:
                    break
        return move

    def observation_(self, position):
        observation = self.environment_memory[
                      position[0] - self.HALF_WINDOW_SIZE:position[0] + self.HALF_WINDOW_SIZE + 1,
                      position[1] - self.HALF_WINDOW_SIZE:position[1] + self.HALF_WINDOW_SIZE + 1]
        return observation.flatten().reshape(1, -1)

    def clip_position(self, position):
        if position[0] <= self.HALF_WINDOW_SIZE:
            position[0] = self.HALF_WINDOW_SIZE
        if position[1] <= self.HALF_WINDOW_SIZE:
            position[1] = self.HALF_WINDOW_SIZE
        if position[0] >= self.plan_width + self.HALF_WINDOW_SIZE - 1:
            position[0] = self.plan_width + self.HALF_WINDOW_SIZE - 1
        if position[1] >= self.plan_width + self.HALF_WINDOW_SIZE - 1:
            position[1] = self.plan_width + self.HALF_WINDOW_SIZE - 1
        return position

    def step(self, action):
        self.count_step += 1

        if self.uniform_step:
            self.step_size = 1
        else:
            self.step_size = np.random.randint(1, 4)

        # 0 left, 1 right, 2 up, 3 down, 4 left_brick, 5 right_brick, 6 up_brick, 7 down_brick

        check = self.check_sur(self.position_memory[-1])
        self.check = check
        # moving: 0 left, 1 right, 2 up, 3 down
        if action == 0 and check[0] == 0:
            step_size = self.move_step([self.position_memory[-1][0], self.position_memory[-1][1]], action,
                                       self.step_size)
            position = [self.position_memory[-1][0], self.position_memory[-1][1] - step_size]
            position = self.clip_position(position)
            self.position_memory.append(position)
        elif action == 1 and check[1] == 0:
            step_size = self.move_step([self.position_memory[-1][0], self.position_memory[-1][1]], action,
                                       self.step_size)
            position = [self.position_memory[-1][0], self.position_memory[-1][1] + step_size]
            position = self.clip_position(position)
            self.position_memory.append(position)
        elif action == 2 and check[2] == 0:
            step_size = self.move_step([self.position_memory[-1][0], self.position_memory[-1][1]], action,
                                       self.step_size)
            position = [self.position_memory[-1][0] + step_size, self.position_memory[-1][1]]
            position = self.clip_position(position)
            self.position_memory.append(position)
        elif action == 3 and check[3] == 0:
            step_size = self.move_step([self.position_memory[-1][0], self.position_memory[-1][1]], action,
                                       self.step_size)
            position = [self.position_memory[-1][0] - step_size, self.position_memory[-1][1]]
            position = self.clip_position(position)
            self.position_memory.append(position)

        # building brick: 4 left_brick, 5 right_brick, 6 up_brick, 7 down_brick
        elif action > 3:
            build_brick = True
            position = self.position_memory[-1]
            self.position_memory.append(position)
            if action == 4 and check[4] == 0:
                self.count_brick += 1
                brick_target = [position[0], position[1] - 1]
                self.environment_memory[brick_target[0], brick_target[1]] += 1.0
            elif action == 5 and check[5] == 0:
                self.count_brick += 1
                brick_target = [position[0], position[1] + 1]
                self.environment_memory[brick_target[0], brick_target[1]] += 1.0
            elif action == 6 and check[7] == 0:
                self.count_brick += 1
                brick_target = [position[0] - 1, position[1]]
                self.environment_memory[brick_target[0], brick_target[1]] += 1.0
            elif action == 7 and check[6] == 0:
                self.count_brick += 1
                brick_target = [position[0] + 1, position[1]]
                self.environment_memory[brick_target[0], brick_target[1]] += 1.0
            else:
                build_brick = False

            done = bool(self.count_brick >= self.total_brick) or check[0:4] == [1, 1, 1, 1]
            if done:
                observation = np.hstack(
                    (self.observation_(position), np.array([[self.count_brick]]), np.array([[self.count_step]])))
                reward = 0.0
                return observation, reward, done
            else:
                if build_brick:
                    reward = self.reward_check(brick_target)
                    observation = np.hstack(
                        (self.observation_(position), np.array([[self.count_brick]]), np.array([[self.count_step]])))
                    return observation, reward, done
        else:
            # illegal_move = useless_move!!
            position = self.position_memory[-1]

        done = bool(self.count_step >= self.total_step) or check[0:4] == [1, 1, 1, 1]
        observation = np.hstack(
            (self.observation_(position), np.array([[self.count_brick]]), np.array([[self.count_step]])))
        reward = 0.0
        return observation, reward, done

    def reward_check(self, position):
        if self.environment_memory[position[0], position[1]] > self.plan[position[0], position[1]]:
            reward = -1.0
        elif self.environment_memory[position[0], position[1]] == self.plan[position[0], position[1]]:
            reward = 10.0
        else:
            reward = 1.0
        return reward

    def plot_3d(self, plan, Env=False):
        if Env:
            z_max = max(max(row) for row in plan)
        else:
            z_max = 1
        X, Y, Z = [], [], []
        for i in range(int(z_max)):
            for j in range(len(plan)):
                for k in range(len(plan[0])):
                    if self.environment_length > plan[j][k] > 0:
                        X.append(k)
                        Y.append(j)
                        Z.append(plan[j][k])
            plan = plan - 1.0
        return X, Y, Z

    def _iou(self):

        plan = self.plan[self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_height,
               self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_width]

        env_memory = self.environment_memory[self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_height,
                     self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_width]

        env_memory_3d = np.zeros((self.plan_height, self.plan_width))
        for i in range(self.plan_height):
            for j in range(self.plan_width):
                if env_memory[i][j] > plan[i][j]:
                    env_memory_3d[i][j] = plan[i][j]
                else:
                    env_memory_3d[i][j] = env_memory[i][j]
        plan_area = self.total_brick
        env_area = self.count_brick
        cross_area = sum(map(sum, env_memory_3d))
        iou = cross_area / (plan_area + env_area - cross_area)
        return iou

        pass



import math
from collections import deque

import gym
import numpy as np


class Env1DStatic(gym.Env):
    def __init__(self, args):
        self.plan_width = 30
        self.plan_height = 20

        self.plan_choose = (args.plan_choose if args.plan_choose is not None else 0)
        self.environment_height = 50
        self.environment_memory = None
        self.count_brick = None
        self.brick_memory = None
        self.HALF_WINDOW_SIZE = args.half_window_size
        self.environment_width = self.plan_width + 2 * self.HALF_WINDOW_SIZE
        self.wall = np.ones((1, 2)) * (-1)
        self.position_memory = None
        self.observation = None
        self.count_step = 0
        self.total_step = 750
        self.plan, self.total_brick = self.create_plan()
        self.action_dim = 3
        self.state_dim = args.half_window_size * 2 + 3
        self.window = args.history_length
        self.uniform_step = args.uniform_step
        self.step_size = 1

        self.features = 7 - 1

    def set_plan_choose(self, plan_choose):
        self.plan_choose = plan_choose

    def get_features(self):
        return self.features

    def normal_distribution(self, x, mean, sigma):
        return np.exp(-1 * ((x - mean) ** 2) / (2 * (sigma ** 2))) / (math.sqrt(2 * np.pi) * sigma)

    def action_space(self):
        return self.action_dim

    def create_plan(self):
        # 0 Sin, 1 Gaussian, 2 Step
        self.one_hot = self.plan_choose

        if self.plan_choose == 0:
            x = np.arange(self.plan_width)
            y = np.round(10 * np.sin(2 * np.pi / self.plan_width * x) + self.plan_height)
        elif self.plan_choose == 1:
            mean1, sigma1 = 0, 3
            x = np.linspace(mean1 - 6 * sigma1, mean1 + 6 * sigma1, 30)
            y = np.round(self.normal_distribution(x, mean1, sigma1) * 100 + 17)
        elif self.plan_choose == 2:
            y = np.ones(30, ) * 15
            y[0:5] = 25
            y[10:15] = 25
            y[20:25] = 25
        else:
            raise ValueError('0: Sin, 1: Gaussian, 2: Step')

        area = sum(y)
        return y, area

    def clip_position(self, position):
        if position <= self.HALF_WINDOW_SIZE:
            position = self.HALF_WINDOW_SIZE
        elif position >= self.plan_width + self.HALF_WINDOW_SIZE - 1:
            position = self.plan_width + self.HALF_WINDOW_SIZE - 1
        return position

    def reset(self):
        self.one_hot = None
        self.plan, self.total_brick = self.create_plan()

        self.environment_memory = np.zeros((1, self.environment_width))
        self.environment_memory[:, :self.HALF_WINDOW_SIZE] = -1
        self.environment_memory[:, -self.HALF_WINDOW_SIZE:] = -1   # e.g.: [-1, -1, 0, 0, 0, 0, 0, 0, -1, -1]

        self.count_brick = 0
        self.count_step = 0

        self.brick_memory = []
        self.position_memory = []
        self.observation = None
        initial_position = self.HALF_WINDOW_SIZE

        self.position_memory.append(initial_position)
        self.brick_memory.append([-1, -1])

        return np.hstack((
            self.environment_memory[:, initial_position - self.HALF_WINDOW_SIZE : initial_position + self.HALF_WINDOW_SIZE + 1],
            np.array([[self.count_brick]]),
            np.array([[self.count_step]])))

    def step(self, action):
        if self.uniform_step:
            self.step_size = 1
        else:
            self.step_size = np.random.randint(1, 4)

        self.count_step += 1
        position = -1

        if action == 0:
            position = self.position_memory[-1] - self.step_size
            position = self.clip_position(position)
            self.position_memory.append(position)
            self.brick_memory.append([-1, -1])

        if action == 1:
            position = self.position_memory[-1] + self.step_size
            position = self.clip_position(position)
            self.position_memory.append(position)
            self.brick_memory.append([-1, -1])

        if action == 2:
            self.count_brick += 1
            position = self.position_memory[-1]
            self.position_memory.append(position)
            self.environment_memory[0, position] += 1
            self.brick_memory.append([position, self.environment_memory[0, position]])

            done = bool(self.count_brick > self.total_brick)

            if done:
                observation = np.hstack(
                    (self.environment_memory[:, position - self.HALF_WINDOW_SIZE: position + self.HALF_WINDOW_SIZE + 1],
                     np.array([[self.count_brick]]),
                     np.array([[self.count_step]])))

                reward = 0.0

                return observation, reward, done
            else:
                done = bool(self.count_step >= self.total_step)
                if self.environment_memory[0, position] > self.plan[position - self.HALF_WINDOW_SIZE]:
                    reward = -1.0
                elif self.environment_memory[0, position] == self.plan[position - self.HALF_WINDOW_SIZE]:
                    reward = 10.0
                else:
                    reward = 1.0

                observation = np.hstack(
                    (self.environment_memory[:, position - self.HALF_WINDOW_SIZE:position + self.HALF_WINDOW_SIZE + 1],
                     np.array([[self.count_brick]]),
                     np.array([[self.count_step]])))

                return observation, reward, done

        done = bool(self.count_step >= self.total_step)
        observation = np.hstack(
            (self.environment_memory[:, position - self.HALF_WINDOW_SIZE:position + self.HALF_WINDOW_SIZE + 1],
             np.array([[self.count_brick]]),
             np.array([[self.count_step]])))
        reward = 0
        return observation, reward, done

    def _iou(self):
        intersection = 0
        union = 0
        for i in range(self.plan_width):
          j = i + self.HALF_WINDOW_SIZE
          intersection += min(self.environment_memory[0, j], self.plan[i])
          union += max(self.environment_memory[0, j], self.plan[i])
        return intersection / union

class Env1DDynamic(gym.Env):
    def __init__(self, args):
        self.plan_width = 30
        self.plan_height = 20

        self.plan_choose = (args.plan_choose if args.plan_choose is not None else 0)
        self.environment_height = 100
        self.environment_memory = None
        self.count_brick = None
        self.brick_memory = None
        self.HALF_WINDOW_SIZE = args.half_window_size
        self.environment_width = self.plan_width + 2 * self.HALF_WINDOW_SIZE
        self.wall = np.ones((1, 2)) * (-1)
        self.position_memory = None
        self.observation = None
        self.count_step = 0
        self.total_step = 750
        self.plan, self.total_brick = self.create_plan()
        self.action_dim = 3
        self.state_dim = args.half_window_size * 2 + 3
        self.window = args.history_length
        self.uniform_step = args.uniform_step

        self.step_size = 1
        self.features = 7 - 1

    def get_features(self):
        return self.features

    def normal_distribution(self, x, mean, sigma):
        return np.exp(-1 * ((x - mean) ** 2) / (2 * (sigma ** 2))) / (math.sqrt(2 * np.pi) * sigma)

    def action_space(self):
        return self.action_dim

    def create_plan(self):
        k_1 = np.random.uniform(3, 12)
        k_2 = np.random.randint(1, 4)
        phase = np.random.uniform(-1, 1) * np.pi
        one_hot = [k_1, k_2, phase]
        self.one_hot = one_hot

        x = np.arange(self.plan_width)
        y = np.round((k_1 * np.sin(2 * np.pi / self.plan_width * (k_2 * x + phase)) + self.plan_height))
        area = sum(y)

        return y, area

    def clip_position(self, position):
        if position <= self.HALF_WINDOW_SIZE:
            position = self.HALF_WINDOW_SIZE
        elif position >= self.plan_width + self.HALF_WINDOW_SIZE - 1:
            position = self.plan_width + self.HALF_WINDOW_SIZE - 1
        return position

    def reset(self):
        self.one_hot = None
        self.plan, self.total_brick = self.create_plan()

        self.environment_memory = np.zeros((1, self.environment_width))
        self.environment_memory[:, :self.HALF_WINDOW_SIZE] = -1
        self.environment_memory[:, -self.HALF_WINDOW_SIZE:] = -1   # e.g.: [-1, -1, 0, 0, 0, 0, 0, 0, -1, -1]

        self.count_brick = 0
        self.count_step = 0

        self.brick_memory = []
        self.position_memory = []
        self.observation = None
        initial_position = self.HALF_WINDOW_SIZE

        self.position_memory.append(initial_position)
        self.brick_memory.append([-1, -1])

        return np.hstack((self.environment_memory[:,
                          initial_position - self.HALF_WINDOW_SIZE:initial_position + self.HALF_WINDOW_SIZE + 1],
                          np.array([[self.count_brick]]), np.array([[self.count_step]])))

    def step(self, action):
        self.count_step += 1

        if self.uniform_step:
            self.step_size = 1
        else:
            self.step_size = np.random.randint(1, 4)

        if action == 0:
            position = self.position_memory[-1] - self.step_size
            position = self.clip_position(position)
            self.position_memory.append(position)
            self.brick_memory.append([-1, -1])

        if action == 1:
            position = self.position_memory[-1] + self.step_size
            position = self.clip_position(position)
            self.position_memory.append(position)
            self.brick_memory.append([-1, -1])

        if action == 2:
            self.count_brick += 1
            position = self.position_memory[-1]
            self.position_memory.append(position)
            self.environment_memory[0, position] += 1
            self.brick_memory.append([position, self.environment_memory[0, position]])

            done = bool(self.count_brick > self.total_brick)

            if done:
                observation = np.hstack(
                    (self.environment_memory[:, position - self.HALF_WINDOW_SIZE: position + self.HALF_WINDOW_SIZE + 1],
                     np.array([[self.count_brick]]),
                     np.array([[self.count_step]])))
                reward = 0.0
                return observation, reward, done
            else:
                done = bool(self.count_step >= self.total_step)
                if self.environment_memory[0, position] > self.plan[position - self.HALF_WINDOW_SIZE]:
                    reward = -1.0
                elif self.environment_memory[0, position] == self.plan[position - self.HALF_WINDOW_SIZE]:
                    reward = 10.0
                else:
                    reward = 1.0

                observation = np.hstack(
                    (self.environment_memory[:, position - self.HALF_WINDOW_SIZE:position + self.HALF_WINDOW_SIZE + 1],
                     np.array([[self.count_brick]]),
                     np.array([[self.count_step]])))

                return observation, reward, done

        done = bool(self.count_step >= self.total_step)
        observation = np.hstack(
            (self.environment_memory[:, position - self.HALF_WINDOW_SIZE:position + self.HALF_WINDOW_SIZE + 1],
             np.array([[self.count_brick]]),
             np.array([[self.count_step]])))
        reward = 0

        return observation, reward, done

    def _iou(self):
        intersection = 0
        union = 0
        for i in range(self.plan_width):
          j = i + self.HALF_WINDOW_SIZE
          intersection += min(self.environment_memory[0, j], self.plan[i])
          union += max(self.environment_memory[0, j], self.plan[i])
        return intersection / union




import os
import csv
import time
import tkinter as tk
import tkinter.messagebox
import tkinter.simpledialog
import tkinter.ttk as ttk
from tkinter import *
import json
import time
import matplotlib.pyplot as plt
from matplotlib.pyplot import MultipleLocator
import numpy as np

from .arguments import get_args
from .env.Env1D import Env1DStatic, Env1DDynamic
from .env.Env2D import Env2DStatic, Env2DDynamic
from .env.Env3D import Env3DStatic, Env3DDynamic


class HumanPlay:
    def __init__(self):

        self.user = tk.simpledialog.askstring('Enter user', 'Enter your first name (all-lowercase):', parent=Tk())
        self.session_id = str(time.time())[-6:]

        # Env
        self.ENV_DICT = {
            1: '1D Static (Plan 1)',
            2: '1D Static (Plan 2)',
            3: '1D Static (Plan 3)',
            4: '1D Dynamic',
            5: '2D Static (Dense Plan)',
            6: '2D Static (Sparse Plan)',
            7: '2D Dynamic (Dense Plan)',
            8: '2D Dynamic (Sparse Plan)',
            9: '3D Static (Dense Plan)',
            10: '3D Static (Sparse Plan)',
            11: '3D Dynamic (Dense Plan)',
            12: '3D Dynamic (Sparse Plan)'
        }
        self.env = None
        self.env_choose = 1  # goes from 1-12 (1D, 2D and 3D plans supported)
        self.args = get_args()
        self.environment_width = None
        self.environment_height = None
        self.window_size = None
        self.max_steps = None
        self.max_bricks = None

        self.brick_action_idx = None
        self.L_action_idx = None
        self.R_action_idx = None
        self.U_action_idx = None
        self.D_action_idx = None
        self.L_brick_action_idx = None
        self.R_brick_action_idx = None
        self.U_brick_action_idx = None
        self.D_brick_action_idx = None

        # State
        self.step_count = 0
        self.brick_count = 0
        self.reward = 0
        self.episode_reward = 0
        self.environment_memory = None
        self.observation = None
        self.done = None

        # GUI
        self.menu_window = Tk()
        self.menu_window.title('Choose Plan')
        self.menu_window.geometry('+200+0')
        self.game_window = None
        self.game_mode = None

        self.plan_window = None
        self.plan_canvas = None

        self.canvas = None
        self.canvas_width = None
        self.canvas_height = None
        self.num_steps_id = None
        self.num_bricks_id = None
        self.num_reward_id = None
        self.num_totalreward_id = None

        self.build_menu_window()
        self.reset_game(self.env_choose)

        #fig = plt.figure(figsize=[5, 5])
        fig = plt.figure()
        self.axe = fig.add_subplot(1, 1, 1, projection='3d')

    def reset_game(self, env_choose):
        self.env_choose = env_choose
        self.step_count = 0
        self.brick_count = 0
        self.reward = 0
        self.episode_reward = 0
        self.done = False
        self.done = False
        self.data = []

        if self.env is not None:
            self.env.close()

        if env_choose <= 4:
            self.display_state = [0 for i in range(5)]
            self.canvas_width = 400
            self.canvas_height = 600

            self.L_action_idx = 0
            self.R_action_idx = 1
            self.brick_action_idx = 2

            if env_choose == 1:
                self.env = Env1DStatic(self.args)
                self.env.set_plan_choose(0)

            elif env_choose == 2:
                self.env = Env1DStatic(self.args)
                self.env.set_plan_choose(1)

            elif env_choose == 3:
                self.env = Env1DStatic(self.args)
                self.env.set_plan_choose(2)

            elif env_choose == 4:
                self.env = Env1DDynamic(self.args)

        elif 4 < env_choose <= 8:
            self.display_state = [0 for i in range(49)]
            self.canvas_width = 400
            self.canvas_height = 400

            self.L_action_idx = 0
            self.R_action_idx = 1
            self.D_action_idx = 2
            self.U_action_idx = 3
            self.brick_action_idx = 4

            if env_choose == 5:
                self.env = Env2DStatic(self.args)
                self.env.set_plan_choose(0)

            elif env_choose == 6:
                self.env = Env2DStatic(self.args)
                self.env.set_plan_choose(1)

            elif env_choose == 7:
                self.env = Env2DDynamic(self.args)
                self.env.set_plan_choose(1)

            elif env_choose == 8:
                self.env = Env2DDynamic(self.args)
                self.env.set_plan_choose(0)

        else:
            self.display_state = [0 for i in range(49)]
            self.canvas_width = 400
            self.canvas_height = 400

            self.L_action_idx = 0
            self.R_action_idx = 1
            self.D_action_idx = 2
            self.U_action_idx = 3
            self.L_brick_action_idx = 4
            self.R_brick_action_idx = 5
            self.U_brick_action_idx = 6
            self.D_brick_action_idx = 7

            if env_choose == 9:
                self.env = Env3DStatic(self.args)
                self.env.set_plan_choose(0)

            elif env_choose == 10:
                self.env = Env3DStatic(self.args)
                self.env.set_plan_choose(1)

            elif env_choose == 11:
                self.env = Env3DDynamic(self.args)
                self.env.set_plan_choose(0)

            elif env_choose == 12:
                self.env = Env3DDynamic(self.args)
                self.env.set_plan_choose(1)

        self.observation = self.env.reset()
        self.reset_game_window()
        self.reset_plan_window()
        self.update_canvas()

    def build_menu_window(self):
        bottom = tk.Frame(self.menu_window).pack(side="bottom")
        right = tk.Frame(self.menu_window).pack(side="right")

        self.game_mode = tk.StringVar()
        self.game_mode_combobox = ttk.Combobox(width=18,
                                               textvariable=self.game_mode,
                                               state="readonly",
                                               justify='center',
                                               values=['Training Mode', 'Evaluation Mode'])
        self.game_mode_combobox.pack()
        self.game_mode_combobox.current(0)
        self.game_mode_combobox.bind('<<ComboboxSelected>>', self.set_mode)

        self.button_env1 = tk.Button(bottom, text="1D Static (Plan 1)", fg="orange",
                                     command=lambda x=1: self.reset_game(x)).pack()
        self.button_env2 = tk.Button(bottom, text="1D Static (Plan 2)", fg="orange",
                                     command=lambda x=2: self.reset_game(x)).pack()
        self.button_env3 = tk.Button(bottom, text="1D Static (Plan 3)", fg="orange",
                                     command=lambda x=3: self.reset_game(x)).pack()
        self.button_env4 = tk.Button(bottom, text="1D Dynamic", fg="green",
                                     command=lambda x=4: self.reset_game(x)).pack()
        self.button_env5 = tk.Button(bottom, text="2D Static (Dense Plan)", fg="purple",
                                     command=lambda x=5: self.reset_game(x)).pack()
        self.button_env6 = tk.Button(bottom, text="2D Static (Sparse Plan)", fg="purple",
                                     command=lambda x=6: self.reset_game(x)).pack()
        self.button_env7 = tk.Button(bottom, text="2D Dynamic (Dense Plan)", fg="red",
                                     command=lambda x=7: self.reset_game(x)).pack()
        self.button_env8 = tk.Button(bottom, text="2D Dynamic (Sparse Plan)", fg="red",
                                     command=lambda x=8: self.reset_game(x)).pack()
        self.button_env9 = tk.Button(bottom, text="3D Static (Dense Plan)", fg="green",
                                     command=lambda x=9: self.reset_game(x)).pack()
        self.button_env10 = tk.Button(bottom, text="3D Static (Sparse Plan)", fg="green",
                                     command=lambda x=10: self.reset_game(x)).pack()
        self.button_env11 = tk.Button(bottom, text="3D Dynamic (Dense Plan)", fg="blue",
                                     command=lambda x=11: self.reset_game(x)).pack()
        self.button_env12 = tk.Button(bottom, text="3D Dynamic (Sparse Plan)", fg="blue",
                                     command=lambda x=12: self.reset_game(x)).pack()
        self.button_env13 = tk.Button(right, text="End Episode", fg="black",
                                     command=self.upon_episode_completion).pack()

    def reset_plan_window(self):
        if self.plan_window is not None:
            self.plan_window.destroy()
            self.plan_window = None

        if self.env_choose in [4, 7, 8, 11, 12]:
            self.plan_window = Tk()
            self.plan_window.geometry('+700+0')
            self.plan_window.title('Dynamic Plan')

            self.plan_canvas = Canvas(
                self.plan_window,
                width=500,
                height=500)

            if self.env_choose == 4:
                self.plan_canvas.create_rectangle(25, 25, 475, 475, outline='gray75', fill='white')
                for i in range(49):
                    self.plan_canvas.create_line(25, 25 + 9 * (i + 1), 475, 25 + 9 * (i + 1), fill='gray85')
                for i in range(29):
                    self.plan_canvas.create_line(25 + 15 * (i + 1), 25, 25 + 15 * (i + 1), 475, fill='gray85')
                for i in range(6):
                    self.plan_canvas.create_text(15, 475 - i * 90, text=str(i * 10), fill='gray60', font=('Arial', 10))
                for i in range(4):
                    self.plan_canvas.create_text(25 + i * 150, 485, text=str(i * 10), fill='gray60', font=('Arial', 10))
                for i, value in enumerate(self.env.plan):
                    self.plan_canvas.create_rectangle(25 + i * 15,
                                                      475,
                                                      25 + (i + 1) * 15,
                                                      475 - 9 * value,
                                                      outline='SeaGreen2',
                                                      fill='SeaGreen2')
            else:
                self.plan_canvas.create_rectangle(30, 30, 470, 470, outline='gray75', fill='white')
                for i in range(20):
                    self.plan_canvas.create_line(30, 30 + 22 * (i + 1), 470, 30 + 22 * (i + 1), fill='gray85')
                    self.plan_canvas.create_line(30 + 22 * (i + 1), 30, 30 + 22 * (i + 1), 470, fill='gray85')
                for i in range(5):
                    self.plan_canvas.create_text(30 + i * 110, 480, text=str(i * 5), fill='gray60', font=('Arial', 10))
                    self.plan_canvas.create_text(20, 470 - i * 110, text=str(i * 5), fill='gray60', font=('Arial', 10))
                for i in range(3, 23):
                    for j in range(3, 23):
                        if self.env.plan[i][j] >= 1:
                            self.plan_canvas.create_rectangle(30 + 22 * (j - 3),
                                                              30 + 22 * (i - 3),
                                                              30 + 22 * (j - 2),
                                                              30 + + 22 * (i - 2),
                                                              outline='SeaGreen2',
                                                              fill='SeaGreen2')
            self.plan_canvas.pack()

    def reset_game_window(self):
        if self.game_window is not None:
            self.game_window.destroy()

        self.game_window = Tk()
        self.game_window.geometry('+250+0')
        self.game_window.title('C.A.M. Human Performance Benchmarking')
        self.game_window.bind("<Key>", self.keypress)
        self.game_window.bind("<Left>", lambda event: self.move('L'))
        self.game_window.bind("<Right>", lambda event: self.move('R'))
        self.game_window.bind("<Up>", lambda event: self.move('U'))
        self.game_window.bind("<Down>", lambda event: self.move('D'))
        self.game_window.bind("<w>", lambda event: self.move('W'))
        self.game_window.bind("<a>", lambda event: self.move('A'))
        self.game_window.bind("<s>", lambda event: self.move('S'))
        self.game_window.bind("<d>", lambda event: self.move('DB'))

        self.canvas = Canvas(
            self.game_window,
            width=self.canvas_width,
            height=self.canvas_height)

        if self.env_choose <= 4:
            self.canvas.create_rectangle(162, 35, 237, 485, outline='gray75')
            for i in range(49):
                self.canvas.create_line(162, 35 + 9 * (i + 1), 237, 35 + 9 * (i + 1), fill='gray85')
            for i in range(4):
                self.canvas.create_line(162 + 15 * (i + 1), 35, 162 + 15 * (i + 1), 485, fill='gray85')
            self.canvas.create_text(152, 485, text=str(0), fill='gray60', font=('Arial', 10))
            self.canvas.create_text(152, 260, text=str(25), fill='gray60', font=('Arial', 10))
            self.canvas.create_text(152, 35, text=str(50), fill='gray60', font=('Arial', 10))
            self.canvas.create_text(200, 497, text='*', fill='red3', font=('Arial', 18))

            self.txt_plan_id = self.canvas.create_text(200, 590, text='plan: ' + self.ENV_DICT[self.env_choose],
                                                       fill='gray60', font=('Arial', 10))

            if self.in_training_mode():
                self.txt_steps_id = self.canvas.create_text(50, 530, text='steps taken',
                                                            fill='gray60', font=('Arial', 12))
                self.txt_bricks_id = self.canvas.create_text(150, 530, text='bricks used',
                                                             fill='gray60', font=('Arial', 12))
                self.num_steps_id = self.canvas.create_text(50, 555, text=str(0),
                                                            fill='gray40', font=('Arial', 18))
                self.num_bricks_id = self.canvas.create_text(150, 555, text=str(0),
                                                             fill='gray40', font=('Arial', 18))
                self.txt_reward_id = self.canvas.create_text(250, 530, text='step reward',
                                                             fill='gray60', font=('Arial', 12))
                self.txt_totalreward_id = self.canvas.create_text(350, 530, text='total reward',
                                                                  fill='gray60', font=('Arial', 12))
                self.num_reward_id = self.canvas.create_text(250, 555, text=str(0),
                                                             fill='gray40', font=('Arial', 18))
                self.num_totalreward_id = self.canvas.create_text(350, 555, text=str(0),
                                                                  fill='gray40', font=('Arial', 18))
            else:
                self.txt_steps_id = self.canvas.create_text(150, 530, text='steps taken',
                                                            fill='gray60', font=('Arial', 12))
                self.txt_bricks_id = self.canvas.create_text(250, 530, text='bricks used',
                                                             fill='gray60', font=('Arial', 12))
                self.num_steps_id = self.canvas.create_text(150, 555, text=str(0),
                                                            fill='gray40', font=('Arial', 18))
                self.num_bricks_id = self.canvas.create_text(250, 555, text=str(0),
                                                             fill='gray40', font=('Arial', 18))
        else:
            for i in range(7):
                for j in range(7):
                    self.display_state[7 * i + j] = self.canvas.create_rectangle(60 + 40 * j,
                                                                                 25 + 40 * i,
                                                                                 60 + 40 * (j + 1),
                                                                                 25 + 40 * (i + 1),
                                                                                 fill='white')
            self.canvas.create_rectangle(60, 25, 340, 305, outline='gray75')
            self.canvas.create_line(60, 65, 340, 65, fill='gray85')
            self.canvas.create_line(60, 105, 340, 105, fill='gray85')
            self.canvas.create_line(60, 145, 340, 145, fill='gray85')
            self.canvas.create_line(60, 185, 340, 185, fill='gray85')
            self.canvas.create_line(60, 225, 340, 225, fill='gray85')
            self.canvas.create_line(60, 265, 340, 265, fill='gray85')
            self.canvas.create_line(100, 25, 100, 305, fill='gray85')
            self.canvas.create_line(140, 25, 140, 305, fill='gray85')
            self.canvas.create_line(180, 25, 180, 305, fill='gray85')
            self.canvas.create_line(220, 25, 220, 305, fill='gray85')
            self.canvas.create_line(260, 25, 260, 305, fill='gray85')
            self.canvas.create_line(300, 25, 300, 305, fill='gray85')
            self.txt_plan_id = self.canvas.create_text(200, 390, text='plan: ' + self.ENV_DICT[self.env_choose],
                                                       fill='gray60', font=('Arial', 10))

            if self.in_training_mode():
                self.txt_steps_id = self.canvas.create_text(50, 330, text='steps taken',
                                                            fill='gray60', font=('Arial', 12))
                self.txt_bricks_id = self.canvas.create_text(150, 330, text='bricks used',
                                                             fill='gray60', font=('Arial', 12))
                self.num_steps_id = self.canvas.create_text(50, 355, text=str(0),
                                                            fill='gray40', font=('Arial', 18))
                self.num_bricks_id = self.canvas.create_text(150, 355, text=str(0),
                                                             fill='gray40', font=('Arial', 18))
                self.txt_reward_id = self.canvas.create_text(250, 330, text='step reward',
                                                             fill='gray60', font=('Arial', 12))
                self.num_reward_id = self.canvas.create_text(250, 355, text=str(0),
                                                             fill='gray40', font=('Arial', 18))
                self.txt_totalreward_id = self.canvas.create_text(350, 330, text='total reward',
                                                                  fill='gray60', font=('Arial', 12))
                self.num_totalreward_id = self.canvas.create_text(350, 355, text=str(0),
                                                                  fill='gray40', font=('Arial', 18))
            else:
                self.txt_steps_id = self.canvas.create_text(150, 330, text='steps taken',
                                                            fill='gray60', font=('Arial', 12))
                self.txt_bricks_id = self.canvas.create_text(250, 330, text='bricks used',
                                                             fill='gray60', font=('Arial', 12))
                self.num_steps_id = self.canvas.create_text(150, 355, text=str(0),
                                                            fill='gray40', font=('Arial', 18))
                self.num_bricks_id = self.canvas.create_text(250, 355, text=str(0),
                                                             fill='gray40', font=('Arial', 18))
            if self.env_choose >= 9:
                self.render3D(self.axe)
        self.canvas.pack()

    def update_canvas(self):
        observation = self.observation.squeeze()
        if self.env_choose <= 4:
            for i in range(5):
                self.canvas.delete(self.display_state[i])
                value = observation[i]
                if value == -1:
                    self.display_state[i] = self.canvas.create_rectangle(162 + 15 * i,
                                                                         35,
                                                                         162 + 15 * (i + 1),
                                                                         485,
                                                                         fill='grey30', outline='gray30')
                elif value > 0:
                    self.display_state[i] = self.canvas.create_rectangle(162 + 15 * i,
                                                                         485,
                                                                         162 + 15 * (i + 1),
                                                                         485 - int(value) * 9,
                                                                         fill='coral2', outline='coral2')
        elif 5 <= self.env_choose <= 8:
            for i in range(7):  # row
                for j in range(7):  # column
                    self.canvas.delete(self.display_state[7 * i + j])
                    value = observation[7 * i + j]
                    if value == -1:
                        self.display_state[7 * i + j] = self.canvas.create_rectangle(60 + 40 * j,
                                                                                     25 + 40 * i,
                                                                                     60 + 40 * (j + 1),
                                                                                     25 + 40 * (i + 1),
                                                                                     fill='grey30', outline='grey30')
                    elif value == 1:
                        self.display_state[7 * i + j] = self.canvas.create_rectangle(60 + 40 * j,
                                                                                     25 + 40 * i,
                                                                                     60 + 40 * (j + 1),
                                                                                     25 + 40 * (i + 1),
                                                                                     fill='coral2', outline='coral2')
            self.canvas.create_text(201, 169, text='*', fill='red3', font=('Arial', 20), anchor=tk.CENTER)

        else:
            for i in range(7):  # row
                for j in range(7):  # column
                    self.canvas.delete(self.display_state[7 * i + j])
                    value = observation[7 * i + j]
                    if value == -1:
                        self.display_state[7 * i + j] = self.canvas.create_rectangle(60 + 40 * j,
                                                                                     25 + 40 * i,
                                                                                     60 + 40 * (j + 1),
                                                                                     25 + 40 * (i + 1),
                                                                                     fill='grey30', outline='grey30')
                    elif value >= 1:
                        self.display_state[7 * i + j] = self.canvas.create_rectangle(60 + 40 * j,
                                                                                     25 + 40 * i,
                                                                                     60 + 40 * (j + 1),
                                                                                     25 + 40 * (i + 1),
                                                                                     fill='coral2', outline='coral2')
            self.canvas.create_text(201, 169, text='*', fill='red3', font=('Arial', 20), anchor=tk.CENTER)
            self.render3D(self.axe)

        self.canvas.itemconfigure(self.num_steps_id, text=str(self.step_count))
        self.canvas.itemconfigure(self.num_bricks_id, text=str(self.brick_count))

        if self.in_training_mode():
            self.canvas.itemconfigure(self.num_reward_id, text=str(self.reward))
            self.canvas.itemconfigure(self.num_totalreward_id, text=str(self.episode_reward))

    def render3D(self, axe):
        axe.clear()

        self.plan_width = self.env.plan_width
        self.plan_height = self.env.plan_height
        self.plan_length = self.env.plan_length
        self.HALF_WINDOW_SIZE = self.env.HALF_WINDOW_SIZE
        self.environment_memory = self.env.environment_memory
        self.position_memory = self.env.position_memory

        plan = self.env.plan[self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_height,
               self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_width]
        env_memory = self.environment_memory[self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_height,
                     self.HALF_WINDOW_SIZE:self.HALF_WINDOW_SIZE + self.plan_width]

        center_x = self.position_memory[-1][1] - self.HALF_WINDOW_SIZE
        center_y = self.position_memory[-1][0] - self.HALF_WINDOW_SIZE

        min_x = 0 if center_x <= 3 else center_x - 3
        max_x = 19 if center_x >= 16 else center_x + 3
        min_y = 0 if center_y <= 3 else center_y - 3
        max_y = 19 if center_y >= 16 else center_y + 3

        axe.set_xlim(min_x, max_x + 1)
        axe.set_ylim(max_y + 1, min_y)
        axe.set_zlim(0, 2 * self.HALF_WINDOW_SIZE + 1)

        view_block = np.zeros((self.plan_height, self.plan_width))
        view_block[min_y:max_y + 1, min_x:max_x + 1] = env_memory[min_y:max_y + 1, min_x:max_x + 1]
        x1, y1, z1 = self.env.plot_3d(view_block, Env=True)

        width = depth = 1
        height = np.zeros_like(z1)
        if x1:
            axe.bar3d(x1, y1, height, width, depth, z1, color='b', shade=True, alpha=0.15, edgecolor="k")
        axe.scatter(self.position_memory[-1][1] - self.HALF_WINDOW_SIZE + 0.5,
                    self.position_memory[-1][0] - self.HALF_WINDOW_SIZE + 0.5, zs=0, marker="*", color='b', s=50)
        axe.view_init(elev=45, azim=315)

        if min_x == 0:
            axe.plot3D(xs = [0, 0], ys = [min_y, max_y + 1], zs = [0, 0])
        if max_x == 19:
            axe.plot3D(xs = [20, 20], ys = [min_y, max_y + 1], zs = [0, 0])
        if min_y == 0:
            axe.plot3D(xs = [min_x, max_x + 1], ys = [0, 0], zs = [0, 0])
        if max_y == 19:
            axe.plot3D(xs = [min_x, max_x + 1], ys = [20, 20], zs = [0, 0])
        x_major_locator = MultipleLocator(1)
        y_major_locator = MultipleLocator(1)
        axe.xaxis.set_major_locator(x_major_locator)
        axe.yaxis.set_major_locator(y_major_locator)
        plt.tick_params(labelsize=1)
        axe.grid()
        plt.draw()
        plt.pause(0.001)

    def upon_episode_completion(self):
        save = tkinter.messagebox.askquestion("Save", "Episode ended - save the result?")

        if not os.path.exists('results'):
            os.makedirs('results')

        tick = time.time()
        filename = 'results/%s.json' % tick
        with open(filename, 'w') as file_obj:
            json.dump(self.data, file_obj)

        if save == 'yes':
            path = 'results/human_results' + '_' + self.user + '_' + self.session_id + '.csv'
            self.log_result(path)
            tkinter.messagebox.showinfo("Saved.", "Episode results appended to:\n" + os.path.abspath(path))
        else:
            tkinter.messagebox.showinfo("Not saved.", "Episode results discarded.")

        self.reset_game(self.env_choose)

    def log_result(self, path):

        if not os.path.exists('results'):
            os.makedirs('results')

        with open(path, 'a', newline='') as log_file:
            schema = ['user', 'env', 'game_mode', 'iou', 'reward', 'num_steps', 'num_bricks']
            writer = csv.DictWriter(log_file, fieldnames=schema)
            writer.writerow({'user': self.user,
                             'env': self.ENV_DICT[self.env_choose],
                             'game_mode': self.game_mode.get(),
                             'iou': self.env._iou(),
                             'reward': self.episode_reward,
                             'num_steps': self.step_count,
                             'num_bricks': self.brick_count})

    def move(self, direction):
        if direction == 'L':
            self.data.append([self.observation.tolist(), self.L_action_idx])
            self.observation, self.reward, self.done = self.env.step(self.L_action_idx)
            self.step_count += 1
            self.after_move()
        elif direction == 'R':
            self.data.append([self.observation.tolist(), self.R_action_idx])
            self.observation, self.reward, self.done = self.env.step(self.R_action_idx)
            self.step_count += 1
            self.after_move()
        elif direction == 'D' and self.env_choose > 4:
            self.data.append([self.observation.tolist(), self.D_action_idx])
            self.observation, self.reward, self.done = self.env.step(self.D_action_idx)
            self.step_count += 1
            self.after_move()
        elif direction == 'U' and self.env_choose > 4:
            self.data.append([self.observation.tolist(), self.U_action_idx])
            self.observation, self.reward, self.done = self.env.step(self.U_action_idx)
            self.step_count += 1
            self.after_move()
        elif direction == 'A' and self.env_choose > 8:
            self.data.append([self.observation.tolist(), self.L_brick_action_idx])
            self.observation, self.reward, self.done = self.env.step(self.L_brick_action_idx)
            self.step_count += 1
            self.brick_count += 1
            self.after_move()
        elif direction == 'DB' and self.env_choose > 8:
            self.data.append([self.observation.tolist(), self.R_brick_action_idx])
            self.observation, self.reward, self.done = self.env.step(self.R_brick_action_idx)
            self.step_count += 1
            self.brick_count += 1
            self.after_move()
        elif direction == 'W' and self.env_choose > 8:
            self.data.append([self.observation.tolist(), self.U_brick_action_idx])
            self.observation, self.reward, self.done = self.env.step(self.U_brick_action_idx)
            self.step_count += 1
            self.brick_count += 1
            self.after_move()
        elif direction == 'S' and self.env_choose > 8:
            self.data.append([self.observation.tolist(), self.D_brick_action_idx])
            self.observation, self.reward, self.done = self.env.step(self.D_brick_action_idx)
            self.step_count += 1
            self.brick_count += 1
            self.after_move()


    def after_move(self):
        self.update_canvas()
        if self.done:
            self.upon_episode_completion()

    def keypress(self, event):
        if event.char == ' ' and self.env_choose <= 8:  # Space bar (drop brick in current location)
            self.data.append([self.observation.tolist(), self.brick_action_idx])
            self.observation, self.reward, self.done = self.env.step(self.brick_action_idx)
            self.episode_reward += self.reward
            self.step_count += 1
            self.brick_count += 1
            self.after_move()


        elif event.char == '\x1b':  # Escape key (stop playing)
            self.upon_episode_completion()
            self.game_window.destroy()
            self.game_window = None

    def in_training_mode(self):
        return self.game_mode.get() == "Training Mode"

    def set_mode(self, event):
        if self.in_training_mode():
            self.reset_game(self.env_choose)
        else:
            self.reset_game(self.env_choose)

    def mainloop(self):
        self.menu_window.mainloop()


from collections import deque

import cv2
import gym
import numpy as np
from matplotlib import patches


class Env2DStatic(gym.Env):
    def __init__(self, args):
        self.plan_width = 20
        self.plan_height = 20

        self.plan_choose = (args.plan_choose if args.plan_choose is not None else 0)

        self.environment_memory = None
        self.count_brick = None
        self.brick_memory = None
        self.HALF_WINDOW_SIZE = 3
        self.environment_width = self.plan_width + 2 * self.HALF_WINDOW_SIZE
        self.environment_height = self.plan_height + 2 * self.HALF_WINDOW_SIZE

        self.position_memory = None
        self.observation = None
        self.count_step = 0
        self.total_step = 600
        self.plan, self.total_brick = self.create_plan()
        self.input_plan = None

        self.action_dim = 5
        self.state_dim = (args.half_window_size * 2 + 1) ** 2 + 2
        self.window = args.history_length
        self.uniform_step = args.uniform_step
        self.step_size = 1

        self.features = 51 - 1

    def set_plan_choose(self, plan_choose):
        self.plan_choose = plan_choose

    def get_features(self):
        return self.features

    def action_space(self):
        return self.action_dim

    def create_plan(self):
        out_radius = 7
        if self.plan_choose == 0:
            in_radius = 0
        elif self.plan_choose == 1:
            out_radius = 8
            in_radius = 7
        else:
            raise ValueError('0: Dense circle, 1: Sparse circle')

        plan = np.zeros((self.environment_height, self.environment_width))
        a = np.array([12.5, 12.5])
        circle = patches.CirclePolygon(a, out_radius)
        circle2 = patches.CirclePolygon(a, in_radius)
        for i in range(len(plan)):
            for j in range(len(plan[0])):
                a = np.array([i, j])
                if circle.contains_point(a) and not circle2.contains_point(a):
                    plan[i][j] = 1

            total_area = sum(sum(plan))

        area=sum(sum(plan))

        return plan, area

    def observation_(self, position):
        observation = self.environment_memory[
                      position[0] - self.HALF_WINDOW_SIZE : position[0] + self.HALF_WINDOW_SIZE + 1,
                      position[1] - self.HALF_WINDOW_SIZE : position[1] + self.HALF_WINDOW_SIZE + 1]

        return observation.flatten().reshape(1, -1)

    def clip_position(self, position):
        if position[0] <= self.HALF_WINDOW_SIZE:
            position[0] = self.HALF_WINDOW_SIZE
        if position[1] <= self.HALF_WINDOW_SIZE:
            position[1] = self.HALF_WINDOW_SIZE
        if position[0] >= self.plan_width+self.HALF_WINDOW_SIZE - 1:
            position[0] = self.plan_width+self.HALF_WINDOW_SIZE - 1
        if position[1] >= self.plan_width+self.HALF_WINDOW_SIZE - 1:
            position[1] = self.plan_width+self.HALF_WINDOW_SIZE -1
        return position

    def reset(self):
        self.plan, self.total_brick = self.create_plan()

        if self.total_brick<30:
            self.total_brick=30

        self.input_plan = self.plan[
                          self.HALF_WINDOW_SIZE : self.HALF_WINDOW_SIZE + self.plan_height,
                          self.HALF_WINDOW_SIZE : self.HALF_WINDOW_SIZE + self.plan_width]
        self.environment_memory = np.zeros((self.environment_height, self.environment_width))
        self.environment_memory[:, :self.HALF_WINDOW_SIZE] = -1
        self.environment_memory[:, -self.HALF_WINDOW_SIZE:] = -1
        self.environment_memory[:self.HALF_WINDOW_SIZE, :] = -1
        self.environment_memory[-self.HALF_WINDOW_SIZE:, :] = -1

        self.count_brick = 0

        self.position_memory = []
        self.observation = None
        self.count_step = 0
        initial_position = [self.HALF_WINDOW_SIZE, self.HALF_WINDOW_SIZE]

        self.position_memory.append(initial_position)
        #
        return np.hstack(
            (self.observation_(initial_position),
             np.array([[self.count_brick]]),
             np.array([[self.count_step]])))

    def step(self, action):
        if self.uniform_step:
            self.step_size = 1
        else:
            self.step_size = np.random.randint(1, 4)

        self.count_step += 1

        ### 0 left, 1 right, 2 up, 3 down
        if action == 0:
            position = [self.position_memory[-1][0], self.position_memory[-1][1] - self.step_size]
            position = self.clip_position(position)
            self.position_memory.append(position)
        if action == 1:
            position = [self.position_memory[-1][0], self.position_memory[-1][1] + self.step_size]
            position = self.clip_position(position)
            self.position_memory.append(position)
        if action == 2:
            position = [self.position_memory[-1][0] + self.step_size, self.position_memory[-1][1]]
            position = self.clip_position(position)
            self.position_memory.append(position)
        if action == 3:
            position = [self.position_memory[-1][0] - self.step_size, self.position_memory[-1][1]]
            position = self.clip_position(position)
            self.position_memory.append(position)

        #######    4   drop
        if action == 4:
            self.count_brick += 1
            position = self.position_memory[-1]
            self.position_memory.append(position)

            self.environment_memory[position[0], position[1]] += 1.0

            done = bool(self.count_brick > self.total_brick)

            if done:
                if self.environment_memory[position[0],position[1]]>1:
                    self.environment_memory[position[0],position[1]]=1.0

                observation = np.hstack(
                    (self.observation_(position),
                     np.array([[self.count_brick]]),
                     np.array([[self.count_step]])))

                reward = 0.0

                return observation, reward, done
            else:
                if self.environment_memory[position[0], position[1]] > self.plan[position[0], position[1]]:
                    reward = 0
                elif self.environment_memory[position[0], position[1]] == self.plan[position[0], position[1]]:
                    reward = 5.0

                if self.environment_memory[position[0], position[1]] > 1.0:
                    self.environment_memory[position[0], position[1]] = 1.0

                observation = np.hstack(
                    (self.observation_(position),
                     np.array([[self.count_brick]]),
                     np.array([[self.count_step]])))

                return observation, reward, done

        done = bool(self.count_step >= self.total_step)
        observation = np.hstack(
            (self.observation_(position),
             np.array([[self.count_brick]]),
             np.array([[self.count_step]])))
        reward = 0

        return observation, reward, done

    def _iou(self):
        component1 = self.plan[
                     self.HALF_WINDOW_SIZE : self.HALF_WINDOW_SIZE + self.plan_height,
                     self.HALF_WINDOW_SIZE : self.HALF_WINDOW_SIZE + self.plan_width]\
            .astype(bool)

        component2 = self.environment_memory[
                     self.HALF_WINDOW_SIZE : self.HALF_WINDOW_SIZE + self.plan_height,
                     self.HALF_WINDOW_SIZE : self.HALF_WINDOW_SIZE + self.plan_width]\
            .astype(bool)

        overlap = component1 * component2 # Logical AND
        union = component1 + component2 # Logical OR

        return overlap.sum()/float(union.sum())

class Env2DDynamic(gym.Env):
    def __init__(self, args):
        self.plan_width = 20
        self.plan_height = 20

        # plan_choose: 0 Dense triangle, 1 Sparse triangle
        self.plan_choose = (args.plan_choose if args.plan_choose is not None else 0)

        self.environment_memory = None
        self.count_brick = None
        self.brick_memory = None
        self.HALF_WINDOW_SIZE = 3
        self.environment_width = self.plan_width + 2 * self.HALF_WINDOW_SIZE
        self.environment_height = self.plan_height + 2 * self.HALF_WINDOW_SIZE

        self.position_memory = None
        self.observation = None
        self.count_step = 0
        self.total_step = 600
        self.plan, self.total_brick = self.create_plan()
        self.input_plan = None

        self.action_dim = 5
        self.state_dim = (args.half_window_size * 2 + 1) ** 2 + 2
        self.window = args.history_length
        self.uniform_step = args.uniform_step
        self.step_size = 1

        self.features = 51 - 1

    def set_plan_choose(self, plan_choose):
        self.plan_choose = plan_choose

    def get_features(self):
        return self.features

    def action_space(self):
        return self.action_dim

    def create_plan(self):
        if not(self.plan_choose == 1 or self.plan_choose == 0 ):
            raise ValueError(' 0: Dense triangle, 1: Sparse triangle')

        plan = np.zeros((self.environment_height, self.environment_width))
        area = [50, 20]
        total_area = 0
        while total_area <= area[self.plan_choose]:
            x = np.random.randint(0, self.plan_width, size=3)
            y = np.random.randint(0, self.plan_height, size=3)

            img_rgb = np.ones((self.plan_height, self.plan_width, 3), np.uint8) * 255
            vertices = np.array([[x[0], y[0]], [x[1], y[1]], [x[2], y[2]]], np.int32)
            pts = vertices.reshape((-1, 1, 2))
            cv2.polylines(img_rgb, [pts], isClosed=True, color=(0, 0, 0))
            if self.plan_choose == 1:
                cv2.fillPoly(img_rgb, [pts], color=(0, 0, 0))

            plan[
            self.HALF_WINDOW_SIZE : self.HALF_WINDOW_SIZE + self.plan_height,
            self.HALF_WINDOW_SIZE : self.HALF_WINDOW_SIZE + self.plan_width] = 1 - img_rgb[:, :, 0] / 255

            total_area = sum(sum(plan))
        return plan, total_area

    def observation_(self, position):
        observation = self.environment_memory[
                      position[0] - self.HALF_WINDOW_SIZE : position[0] + self.HALF_WINDOW_SIZE + 1,
                      position[1] - self.HALF_WINDOW_SIZE : position[1] + self.HALF_WINDOW_SIZE + 1]

        return observation.flatten().reshape(1, -1)

    def clip_position(self, position):
        if position[0] <= self.HALF_WINDOW_SIZE:
            position[0] = self.HALF_WINDOW_SIZE
        if position[1] <= self.HALF_WINDOW_SIZE:
            position[1] = self.HALF_WINDOW_SIZE
        if position[0] >= self.plan_width+self.HALF_WINDOW_SIZE - 1:
            position[0] = self.plan_width+self.HALF_WINDOW_SIZE - 1
        if position[1] >= self.plan_width+self.HALF_WINDOW_SIZE - 1:
            position[1] = self.plan_width+self.HALF_WINDOW_SIZE -1
        return position

    def reset(self):
        self.plan, self.total_brick = self.create_plan()

        if self.total_brick<30:
            self.total_brick=30

        self.input_plan = self.plan[
                          self.HALF_WINDOW_SIZE : self.HALF_WINDOW_SIZE + self.plan_height,
                          self.HALF_WINDOW_SIZE : self.HALF_WINDOW_SIZE + self.plan_width]
        self.environment_memory = np.zeros((self.environment_height, self.environment_width))
        self.environment_memory[:, :self.HALF_WINDOW_SIZE] = -1
        self.environment_memory[:, -self.HALF_WINDOW_SIZE:] = -1
        self.environment_memory[:self.HALF_WINDOW_SIZE, :] = -1
        self.environment_memory[-self.HALF_WINDOW_SIZE:, :] = -1

        self.count_brick = 0

        self.position_memory = []
        self.observation = None
        self.count_step = 0
        initial_position = [self.HALF_WINDOW_SIZE, self.HALF_WINDOW_SIZE]

        self.position_memory.append(initial_position)
        #
        return np.hstack(
            (self.observation_(initial_position),
             np.array([[self.count_brick]]),
             np.array([[self.count_step]])))

    def step(self, action):
        if self.uniform_step:
            self.step_size = 1
        else:
            self.step_size = np.random.randint(1, 4)

        self.count_step += 1

        ### 0 left, 1 right, 2 up, 3 down
        if action == 0:
            position = [self.position_memory[-1][0], self.position_memory[-1][1] - self.step_size]
            position = self.clip_position(position)
            self.position_memory.append(position)
        if action == 1:
            position = [self.position_memory[-1][0], self.position_memory[-1][1] + self.step_size]
            position = self.clip_position(position)
            self.position_memory.append(position)
        if action == 2:
            position = [self.position_memory[-1][0] + self.step_size, self.position_memory[-1][1]]
            position = self.clip_position(position)
            self.position_memory.append(position)
        if action == 3:
            position = [self.position_memory[-1][0] - self.step_size, self.position_memory[-1][1]]
            position = self.clip_position(position)
            self.position_memory.append(position)

        #######    4   drop
        if action == 4:
            self.count_brick += 1
            position = self.position_memory[-1]
            self.position_memory.append(position)

            self.environment_memory[position[0], position[1]] += 1.0

            done = bool(self.count_brick > self.total_brick)

            if done:
                if self.environment_memory[position[0],position[1]]>1:
                    self.environment_memory[position[0],position[1]]=1.0

                observation = np.hstack(
                    (self.observation_(position),
                     np.array([[self.count_brick]]),
                     np.array([[self.count_step]])))

                reward = 0.0

                return observation, reward, done
            else:
                if self.environment_memory[position[0], position[1]] > self.plan[position[0], position[1]]:
                    reward = 0
                elif self.environment_memory[position[0], position[1]] == self.plan[position[0], position[1]]:
                    reward = 5.0

                if self.environment_memory[position[0], position[1]] > 1.0:
                    self.environment_memory[position[0], position[1]] = 1.0

                observation = np.hstack(
                    (self.observation_(position),
                     np.array([[self.count_brick]]),
                     np.array([[self.count_step]])))

                return observation, reward, done

        done = bool(self.count_step >= self.total_step)
        observation = np.hstack(
            (self.observation_(position),
             np.array([[self.count_brick]]),
             np.array([[self.count_step]])))
        reward = 0

        return observation, reward, done

    def _iou(self):
        component1 = self.plan[
                     self.HALF_WINDOW_SIZE : self.HALF_WINDOW_SIZE + self.plan_height,
                     self.HALF_WINDOW_SIZE : self.HALF_WINDOW_SIZE + self.plan_width]\
            .astype(bool)

        component2 = self.environment_memory[
                     self.HALF_WINDOW_SIZE : self.HALF_WINDOW_SIZE + self.plan_height,
                     self.HALF_WINDOW_SIZE : self.HALF_WINDOW_SIZE + self.plan_width]\
            .astype(bool)

        overlap = component1 * component2 # Logical AND
        union = component1 + component2 # Logical OR

        return overlap.sum()/float(union.sum())


from setuptools import setup, find_packages
setup(name='mobileprint-2',
      version='0.0.2',
      description='GUI for 1D，2D and 3D environments',
      author='Dennis',
      author_email='dennis2016@163.com',
      requires= ['numpy','matplotlib'], # 定义依赖哪些模块
      packages=find_packages(),  # 系统自动从当前目录开始找包
      # 如果有的文件不用打包，则只能指定需要打包的文件
      #packages=['代码1','代码2','__init__']  #指定目录中需要打包的py文件，注意不要.py后缀
      license="MIT"
      )

'''
name : 打包后包的文件名
version : 版本号
author : 作者
author_email : 作者的邮箱
py_modules : 要打包的.py文件
packages: 打包的python文件夹
include_package_data : 项目里会有一些非py文件,比如html和js等,这时候就要靠include_package_data 和 package_data 来指定了。package_data:一般写成{‘your_package_name’: [“files”]}, include_package_data还没完,还需要修改MANIFEST.in文件.MANIFEST.in文件的语法为: include xxx/xxx/xxx/.ini/(所有以.ini结尾的文件,也可以直接指定文件名)
license : 支持的开源协议
description : 对项目简短的一个形容
ext_modules : 是一个包含Extension实例的列表,Extension的定义也有一些参数。
ext_package : 定义extension的相对路径
requires : 定义依赖哪些模块
provides : 定义可以为哪些模块提供依赖
data_files :指定其他的一些文件(如配置文件),规定了哪些文件被安装到哪些目录中。如果目录名是相对路径,则是相对于sys.prefix或sys.exec_prefix的路径。如果没有提供模板,会被添加到MANIFEST文件中。
'''


import numpy as np
import random

from collections import deque

class ReplayBuffer(object):
    def __init__(self, capacity):
        self.buffer = deque(maxlen=capacity)
    
    def push(self, state, action, reward, next_state, done):
        state = np.expand_dims(state, 0)
        next_state = np.expand_dims(next_state, 0)
            
        self.buffer.append((state, action, reward, next_state, done))
    
    def sample(self, batch_size):
        state, action, reward, next_state, done = zip(*random.sample(self.buffer, batch_size))
        return np.concatenate(state), action, reward, np.concatenate(next_state), done
    
    def __len__(self):
        return len(self.buffer)

class NaivePrioritizedBuffer(object):
    def __init__(self, capacity, prob_alpha=0.6):
        self.prob_alpha = prob_alpha
        self.capacity   = capacity
        self.buffer     = []
        self.pos        = 0
        self.priorities = np.zeros((capacity,), dtype=np.float32)
    
    def push(self, state, action, reward, next_state, done):
        assert state.ndim == next_state.ndim
        state      = np.expand_dims(state, 0)
        next_state = np.expand_dims(next_state, 0)
        
        max_prio = self.priorities.max() if self.buffer else 1.0
        
        if len(self.buffer) < self.capacity:
            self.buffer.append((state, action, reward, next_state, done))
        else:
            self.buffer[self.pos] = (state, action, reward, next_state, done)
        
        self.priorities[self.pos] = max_prio
        self.pos = (self.pos + 1) % self.capacity
    
    def sample(self, batch_size, beta=0.4):
        if len(self.buffer) == self.capacity:
            prios = self.priorities
        else:
            prios = self.priorities[:self.pos]
        
        probs  = prios ** self.prob_alpha
        probs /= probs.sum()
        
        indices = np.random.choice(len(self.buffer), batch_size, p=probs)
        samples = [self.buffer[idx] for idx in indices]
        
        total    = len(self.buffer)
        weights  = (total * probs[indices]) ** (-beta)
        weights /= weights.max()
        weights  = np.array(weights, dtype=np.float32)
        
        batch       = list(zip(*samples))
        states      = np.concatenate(batch[0])
        actions     = batch[1]
        rewards     = batch[2]
        next_states = np.concatenate(batch[3])
        dones       = batch[4]
        
        return states, actions, rewards, next_states, dones, weights, indices
    
    def update_priorities(self, batch_indices, batch_priorities):
        for idx, prio in zip(batch_indices, batch_priorities):
            self.priorities[idx] = prio

    def __len__(self):
        return len(self.buffer)


import argparse

def get_args():
    parser = argparse.ArgumentParser(description='DQN')

    # Basic Arguments
    parser.add_argument('--results-path', type=str, default='results/',
                        help='Path to folder containing results.')
    parser.add_argument('--extra-note', type=str, default='dim6only',
                        help='any other changes to note about the current experiment.')
    parser.add_argument('--seed', type=int, default=123,
                        help='Random seed')
    parser.add_argument('--batch-size', type=int, default=32,
                        help='Batch size')
    parser.add_argument('--no-cuda', action='store_true', default=False,
                        help='disables CUDA training')

    # Training Arguments
    parser.add_argument('--max-frames', type=int, default=1500000, metavar='STEPS',
                        help='Number of frames to train')
    parser.add_argument('--buffer-size', type=int, default=10000, metavar='CAPACITY',
                        help='Maximum memory buffer size')
    parser.add_argument('--update-target', type=int, default=750, metavar='STEPS',
                        help='Interval of target network update')
    parser.add_argument('--train-freq', type=int, default=1, metavar='STEPS',
                        help='Number of steps between optimization step')
    parser.add_argument('--gamma', type=float, default=0.9, metavar='γ',
                        help='Discount factor')
    parser.add_argument('--learning-start', type=int, default=1000, metavar='N',
                        help='How many steps of the model to collect transitions for before learning starts')
    parser.add_argument('--eps_start', type=float, default=1.0,
                        help='Start value of epsilon')
    parser.add_argument('--eps_final', type=float, default=0.01,
                        help='Final value of epsilon')
    parser.add_argument('--eps_decay', type=int, default=30000,
                        help='Adjustment parameter for epsilon')
    parser.add_argument('--half_window_size', type=int, default=2,
                        help='Half the size of partially-observable window.')
    parser.add_argument('--max-episode-length', type=int, default=int(750), metavar='LENGTH',
                        help='Max episode length in game frames (0 to disable)')
    parser.add_argument('--history-length', type=int, default=1, metavar='T',
                        help='Number of consecutive states processed')
    parser.add_argument('--evaluation-episodes', type=int, default=20, metavar='N',
                        help='Number of evaluation episodes to average over')

    # Algorithm Arguments
    parser.add_argument('--double', action='store_true',
                        help='Enable Double-Q Learning')
    parser.add_argument('--dueling', action='store_true',
                        help='Enable Dueling Network')
    parser.add_argument('--noisy', action='store_true',
                        help='Enable Noisy Network')
    parser.add_argument('--prioritized-replay', action='store_true',
                        help='enable prioritized experience replay')
    parser.add_argument('--c51', action='store_true',
                        help='enable categorical dqn')
    parser.add_argument('--multi-step', type=int, default=3,
                        help='N-Step Learning')
    parser.add_argument('--Vmin', type=int, default=-5,
                        help='Minimum value of support for c51')
    parser.add_argument('--Vmax', type=int, default=35,
                        help='Maximum value of support for c51')
    parser.add_argument('--num-atoms', type=int, default=81,
                        help='Number of atom for c51')
    parser.add_argument('--alpha', type=float, default=0.5,
                        help='Alpha value for prioritized replay')
    parser.add_argument('--beta-start', type=float, default=0.4,
                        help='Start value of beta for prioritized replay')
    parser.add_argument('--beta-frames', type=int, default=100000,
                        help='End frame of beta schedule for prioritized replay')
    parser.add_argument('--sigma-init', type=float, default=0.1,
                        help='Sigma initialization value for NoisyNet')

    # Environment Arguments
    parser.add_argument('--env', type=str, default='1DStatic',
                        help='Environment Name')
    parser.add_argument('--episode-life', type=int, default=1,
                        help='Whether env has episode life(1) or not(0)')
    parser.add_argument('--clip-rewards', type=int, default=0,
                        help='Whether env clip rewards(1) or not(0)')
    parser.add_argument('--frame-stack', type=int, default=0,
                        help='Whether env stacks frame(1) or not(0)')
    parser.add_argument('--scale', type=int, default=0,
                        help='Whether env scales(1) or not(0)')
    parser.add_argument('--uniform_step', action='store_true',
                        help='Enable Double-Q Learning')
    parser.add_argument('--plan-choose', type=int, default=None,
                        help='Specifies which plan to train/test on.')

    # Evaluation Arguments
    parser.add_argument('--load-model', type=str, default=None,
                        help='Pretrained model name to load (state dict)')
    parser.add_argument('--save-model', type=str, default='model',
                        help='Pretrained model name to save (state dict)')
    parser.add_argument('--evaluate', action='store_true',
                        help='Evaluate only')
    parser.add_argument('--render', action='store_true',
                        help='Render evaluation agent')
    parser.add_argument('--evaluation-interval', type=int, default=750,
                        help='Frames for evaluation interval')

    # Optimization Arguments
    parser.add_argument('--lr', type=float, default=.00005, metavar='η',
                        help='Learning rate')

    args = parser.parse_args()

    return args
